import asyncio
from typing import Callable, TypeVar, Optional

from adhesive.model.AdhesiveProcess import AdhesiveProcess
from adhesive.consoleui.ConsoleUserTaskProvider import ConsoleUserTaskProvider
from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.steps.AdhesiveTask import AdhesiveTask
from adhesive.steps.AdhesiveUserTask import AdhesiveUserTask
from adhesive.xml.bpmn import read_bpmn_file

T = TypeVar('T')
process = AdhesiveProcess('_root')


def task(*task_names: str) -> Callable[..., Callable[..., T]]:
    def wrapper_builder(f: Callable[..., T]) -> Callable[..., T]:
        process.steps.append(AdhesiveTask(f, *task_names))
        return f

    return wrapper_builder


def usertask(*task_names: str) -> Callable[..., Callable[..., T]]:
    def wrapper_builder(f: Callable[..., T]) -> Callable[..., T]:
        process.steps.append(AdhesiveUserTask(f, *task_names))
        return f

    return wrapper_builder


def bpmn_build(file_name: str,
               ut_provider: Optional['UserTaskProvider'] = None,
               wait_tasks: bool=True):
    """ Start a build that was described in BPMN """
    process.workflow = read_bpmn_file(file_name)

    if ut_provider is None:
        ut_provider = ConsoleUserTaskProvider()

    fn = WorkflowExecutor(
        process,
        ut_provider=ut_provider,
        wait_tasks=wait_tasks).execute()

    asyncio.get_event_loop().run_until_complete(fn)


def bpmn_build_async(
        file_name: str,
        ut_provider: Optional['UserTaskProvider'] = None,
        wait_tasks: bool=True):
    """ Start a build that was described in BPMN """
    process.workflow = read_bpmn_file(file_name)

    if ut_provider is None:
        ut_provider = ConsoleUserTaskProvider()

    return WorkflowExecutor(
        process,
        ut_provider=ut_provider,
        wait_tasks=wait_tasks).execute()


from adhesive.model.UserTaskProvider import UserTaskProvider



from abc import ABC, abstractmethod
from typing import Optional, List, Dict, Any, Iterable, Union, Tuple


class UiBuilderApi(ABC):
    @abstractmethod
    def add_input_text(self,
                       name: str,
                       title: Optional[str] = None,
                       value: str = '') -> None:
        pass

    @abstractmethod
    def add_input_password(self,
                           name: str,
                           title: Optional[str] = None,
                           value: str = '') -> None:
        pass

    @abstractmethod
    def add_combobox(self,
                     name: str,
                     title: Optional[str] = None,
                     value: Optional[str]=None,
                     values: Optional[Iterable[Union[Tuple[str, str], str]]]=None) -> None:
        pass

    @abstractmethod
    def add_checkbox_group(
            self,
            name: str,
            title: Optional[str]=None,
            value: Optional[Iterable[str]]=None,
            values: Optional[Iterable[Union[Tuple[str, str], str]]]=None) -> None:
        pass

    @abstractmethod
    def add_radio_group(self,
                        name: str,
                        title: Optional[str]=None,
                        value: Optional[str]=None,
                        values: Optional[List[Any]]=None) -> None:
        pass

    @abstractmethod
    def add_default_button(self,
                           name: str,
                           title: Optional[str]=None) -> None:
        pass



from typing import List, cast

from adhesive.graph.Edge import Edge
from adhesive.graph.ExclusiveGateway import ExclusiveGateway
from adhesive.graph.Gateway import Gateway
from adhesive.graph.BaseTask import BaseTask
from adhesive.graph.Workflow import Workflow
from adhesive.model.ActiveEvent import ActiveEvent


class GatewayController:
    @staticmethod
    def compute_outgoing_edges(workflow, event) -> List[Edge]:
        if isinstance(event.task, ExclusiveGateway):
            gateway = cast(Gateway, event.task)
            outgoing_edges = GatewayController.route_single_output(
                workflow, gateway, event)
        else:
            outgoing_edges = GatewayController.route_all_outputs(
                workflow, event.task, event)

        return outgoing_edges

    @staticmethod
    def route_single_output(
            workflow: Workflow,
            gateway: Gateway,
            event: ActiveEvent) -> List[Edge]:

        default_edge = None
        result_edge = None

        edges = workflow.get_outgoing_edges(gateway.id)

        for edge in edges:
            if not edge.condition:
                if default_edge is not None:
                    raise Exception(f"Duplicate default edge for gateway {gateway.id}.")

                default_edge = edge
                continue

            if eval(edge.condition, globals(), event.context.as_mapping()):
                if result_edge is not None:
                    raise Exception(f"Duplicate output edge for gateway {gateway.id}")

                result_edge = edge
                continue

        if result_edge is None and default_edge is not None:
            result_edge = default_edge

        if not result_edge:
            raise Exception(f"No branch matches on gateway {gateway.id}")

        return [result_edge]

    @staticmethod
    def route_all_outputs(
            workflow: Workflow,
            task: BaseTask,
            event: ActiveEvent) -> List[Edge]:

        result_edges = []
        edges = workflow.get_outgoing_edges(task.id)

        for edge in edges:
            # if we have no condition on the edge, we create an event for it
            if edge.condition and not eval(edge.condition, globals(), event.context.as_mapping()):
                continue

            result_edges.append(edge)

        return result_edges



from typing import cast

from adhesive.graph.ScriptTask import ScriptTask
from adhesive.model.ActiveEvent import ActiveEvent
from adhesive.steps.WorkflowContext import WorkflowContext


def call_script_task(event: ActiveEvent) -> WorkflowContext:
    exec(
        cast(ScriptTask, event.task).script,
        {},                          # globals
        {
             "context": event.context,
             "loop": event.context.loop,
             "data": event.context.data,
        })  # locals

    return event.context



import unittest

from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.xml.bpmn import read_bpmn_file
from test.adhesive.steps.ui_provider import TestUserTaskProvider
from .test_tasks import adhesive, _async


class TestUserTaskBasic(unittest.TestCase):
    def test_link_back_execution(self):
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/user-task.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process,
                                             ut_provider=TestUserTaskProvider())
        data = _async(workflow_executor.execute())

        self.assertEqual("OK", data.OK)
        self.assertEqual("Cancel", data.Cancel)
        self.assertEqual("branch", data.branch)
        self.assertEqual("12.0", data.version)
        self.assertEqual("password", data.password)
        self.assertEqual(('integration',), data.run_tests)

        self.assertFalse(workflow_executor.events)


if __name__ == '__main__':
    unittest.main()



from typing import Optional, List, Iterable, Union, Tuple, Any

from adhesive.model.ActiveEvent import ActiveEvent
from adhesive.model.UiBuilderApi import UiBuilderApi
from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.model.UserTaskProvider import UserTaskProvider
from adhesive.steps.WorkflowContext import WorkflowContext


class UIBuilder(UiBuilderApi):
    def __init__(self,
                 context: WorkflowContext):
        self.context = context

    def add_input_text(self,
                       name: str,
                       title: Optional[str] = None,
                       value: str = '') -> None:
        self.context.data[name] = name

    def add_input_password(self,
                           name: str,
                           title: Optional[str] = None,
                           value: str = '') -> None:
        self.context.data[name] = name

    def add_combobox(self,
                     name: str,
                     title: Optional[str] = None,
                     value: Optional[str]=None,
                     values: Optional[Iterable[Union[Tuple[str, str], str]]]=None) -> None:
        self.context.data[name] = values[0]

    def add_checkbox_group(
            self,
            name: str,
            title: Optional[str]=None,
            value: Optional[Iterable[str]]=None,
            values: Optional[Iterable[Union[Tuple[str, str], str]]]=None) -> None:
        self.context.data[name] = value

    def add_radio_group(self,
                        name: str,
                        title: Optional[str]=None,
                        value: Optional[str]=None,
                        values: Optional[List[Any]]=None) -> None:
        pass

    def add_default_button(self,
                           name: str,
                           title: Optional[str]=None) -> None:
        self.context.data[name] = name


class TestUserTaskProvider(UserTaskProvider):
    def __init__(self):
        super(TestUserTaskProvider, self).__init__()

    def register_event(self,
                       executor: WorkflowExecutor,
                       event: ActiveEvent) -> None:
        try:
            ui = UIBuilder(event.context)

            adhesive_task = executor.tasks_impl[event.task.id]
            context = adhesive_task.invoke_user_task(event.context, ui)

            event.future.set_result(context)
        except Exception as e:
            event.future.set_exception(e)




import uuid
from typing import Callable, Any, Optional

from adhesive.graph.BaseTask import BaseTask
from adhesive.model.ActiveEvent import ActiveEvent


class WorkflowLoop:
    """
    Holds the current looping information.
    """
    def __init__(self,
                 loop_id: str,
                 parent_loop: Optional['WorkflowLoop'],
                 task: BaseTask,
                 item: Any,
                 index: int) -> None:
        self.loop_id = loop_id
        self._task = task
        self._key = item
        self._value = item
        self._index = index
        self.parent_loop = parent_loop

    @property
    def task(self) -> BaseTask:
        return self._task

    @property
    def key(self) -> Any:
        return self._key

    @property
    def value(self) -> Any:
        return self._value

    @property
    def index(self) -> int:
        return self._index

    @staticmethod
    def create_loop(event: 'ActiveEvent',
                    clone_event: Callable[['ActiveEvent', 'BaseTask'], 'ActiveEvent']) -> None:
        expression = event.task.loop.loop_expression

        result = eval(expression, {}, {
            "context": event.context,
            "data": event.context.data,
            "loop": event.context.loop,
        })

        if not result:
            return

        index = 0
        for item in result:
            new_event = clone_event(event, event.task)
            loop_id = str(uuid.uuid4())

            parent_loop = new_event.context.loop
            new_event.context.loop = WorkflowLoop(
                loop_id,
                parent_loop,
                event.task,
                item,
                index)

            # if we're iterating over a map, we're going to store the
            # values as well.
            if isinstance(result, dict):
                new_event.context.loop._value = result[item]

            new_event.context.update_title()

            index += 1



import logging
import sys
from concurrent.futures import Future
from typing import Set, Optional, Dict, TypeVar, Any, List, Tuple

from adhesive.graph.BoundaryEvent import BoundaryEvent
from adhesive.graph.Gateway import Gateway, NonWaitingGateway, WaitingGateway
from adhesive.graph.ScriptTask import ScriptTask
from adhesive.graph.Task import Task
from adhesive.graph.UserTask import UserTask
from adhesive.model.ActiveEventStateMachine import ActiveEventState
from adhesive.model.GatewayController import GatewayController
from adhesive.model.WorkflowExecutorConfig import WorkflowExecutorConfig
from adhesive.model.generate_methods import display_unmatched_tasks
from adhesive.steps.AdhesiveBaseTask import AdhesiveBaseTask
from adhesive.steps.WorkflowContext import WorkflowContext
from adhesive.steps.WorkflowData import WorkflowData
from adhesive.steps.WorkflowLoop import WorkflowLoop
from adhesive.steps.call_script_task import call_script_task

T = TypeVar('T')

import concurrent.futures

from adhesive.graph.EndEvent import EndEvent
from adhesive.graph.StartEvent import StartEvent
from adhesive.graph.SubProcess import SubProcess
from adhesive.graph.BaseTask import BaseTask
from adhesive.graph.Workflow import Workflow
from adhesive.steps.AdhesiveTask import AdhesiveTask

from .ActiveEvent import ActiveEvent
from .AdhesiveProcess import AdhesiveProcess

LOG = logging.getLogger(__name__)

DONE_STATES = {
    ActiveEventState.DONE_CHECK,
    ActiveEventState.DONE_END_TASK,
    ActiveEventState.DONE,
}

ACTIVE_STATES = {
    ActiveEventState.NEW,
    ActiveEventState.PROCESSING,
    ActiveEventState.WAITING,
    ActiveEventState.RUNNING,
    ActiveEventState.ERROR,
    ActiveEventState.ROUTING,
}

# When waiting for predecessors it only makes sense to collapse events
# into ActiveEvents only when the event is not already running.
PRE_RUN_STATES = {
    ActiveEventState.NEW,
    ActiveEventState.PROCESSING,
    ActiveEventState.WAITING,
}


def is_predecessor(event, e) -> bool:
    if e.state.state not in ACTIVE_STATES:
        return False

    if e.task == event.task:
        return False

    if event.context.loop:
        # if we are in a loop and the other predecessor is inside
        # a loop of its own, we need to check if it's in the same
        # loop as ours
        if e.task.loop:
            return parent_loop_id(e) == loop_id(event)

        # we check if we are in the same loop. each iteration
        # will have its own loop id.
        if loop_id(e) != loop_id(event):
            return False

    return True


def parent_loop_id(e: ActiveEvent) -> Optional[str]:
    if not e.context.loop:
        return None

    if not e.context.loop.parent_loop:
        return None

    return e.context.loop.parent_loop.loop_id


def loop_id(e: ActiveEvent) -> Optional[str]:
    if not e.context.loop:
        return None

    return e.context.loop.loop_id


class WorkflowExecutor:
    """
    An executor of AdhesiveProcesses.
    """
    pool = concurrent.futures.ProcessPoolExecutor()

    def __init__(self,
                 process: AdhesiveProcess,
                 ut_provider: Optional['UserTaskProvider'] = None,
                 wait_tasks: bool = True) -> None:
        self.process = process
        self.tasks_impl: Dict[str, AdhesiveTask] = dict()

        # A dictionary of events that are currently active. This is just to find out
        # the parent of an event, since from it we can also derive the current parent
        # workflow.
        self.events: Dict[str, ActiveEvent] = dict()
        self.futures: Dict[Any, str] = dict()
        self.ut_provider = ut_provider

        self.config = WorkflowExecutorConfig(wait_tasks=wait_tasks)

    async def execute(self) -> WorkflowData:
        """
        Execute the current events. This will ensure new events are
        generating for forked events.
        """
        workflow = self.process.workflow
        self.tasks_impl: Dict[str, AdhesiveTask] = dict()

        self._validate_tasks(workflow)

        workflow_context = WorkflowContext(workflow)
        fake_event = ActiveEvent(parent_id=None, context=workflow_context)
        fake_event.id = None

        root_event = self.clone_event(fake_event, workflow)

        def raise_exception(_ev):
            raise _ev.data

        root_event.state.after_enter(ActiveEventState.ERROR, raise_exception)

        await self.execute_workflow()

        return root_event.context.data

    async def execute_workflow(self) -> None:
        """
        Process the events in a workflow until no more events are available.
        For example an event is the start of the workflow. The events are
        then creating futures (i.e. actual stuff that's being processed)
        that in turn might generate new events.
        :return:
        """
        while self.events:
            pending_events = filter(lambda e: e.state.state == ActiveEventState.NEW,
                                    self.events.values())

            # except for new processes, and processes that are running with futures
            # the states should transition automatically.
            processed_events = list(pending_events)
            for pending_event in processed_events:
                pending_event.state.process()

            done_futures, not_done_futures = concurrent.futures.wait(
                self.futures.keys(),
                return_when=concurrent.futures.FIRST_COMPLETED,
                timeout=5)

            for future in done_futures:
                event_id = self.futures[future]
                try:
                    context = future.result()
                    self.events[event_id].state.route(context)
                except Exception as e:
                    self.events[event_id].state.error(e)

    def register_event(self,
                       event: ActiveEvent) -> ActiveEvent:
        """
        Register the event into a big map for finding, so we can access it later
        for parents and what not, without serializing the event graph to
        the subprocesses.

        :param event:
        :return:
        """
        if event.id in self.events:
            raise Exception(f"Event {event.id} is already registered as "
                            f"{self.events[event.id]}. Got a new request to register "
                            f"it as {event}.")

        LOG.debug(f"Register {event}")

        self.events[event.id] = event

        return event

    def unregister_event(self,
                         event: ActiveEvent) -> None:
        if event.id not in self.events:
            raise Exception(f"{event} not found in events. Either the event was "
                            f"already terminated, either it was not registered.")

        LOG.debug(f"Unregister {event}")

        del self.events[event.id]

    def get_parent(self,
                   event_id: str) -> ActiveEvent:
        """
        Find the parent of an event by looking at the event ids.
        :param event_id:
        :return:
        """
        event = self.events[event_id]
        parent = self.events[event.parent_id]

        return parent

    def _validate_tasks(self,
                        workflow: Workflow) -> None:
        """
        Recursively traverse the graph, and print to the user if it needs to implement
        some tasks.

        :param workflow:
        :return:
        """
        unmatched_tasks: Set[BaseTask] = set()

        for task_id, task in workflow.tasks.items():
            if isinstance(task, SubProcess):
                self._validate_tasks(task)
                continue

            # gateways don't have associated tasks with them.
            if isinstance(task, StartEvent) or \
                    isinstance(task, EndEvent) or \
                    isinstance(task, Gateway) or \
                    isinstance(task, BoundaryEvent):
                continue

            if isinstance(task, ScriptTask):
                if task.language in ("python", "text/python"):
                    continue

                raise Exception(f"Unknown script task language: {task.language}. Only python and "
                                f"text/python are supported.")

            adhesive_step = self._match_task(task)

            self.tasks_impl[task_id] = adhesive_step

            if not adhesive_step:
                unmatched_tasks.add(task)

        if unmatched_tasks:
            display_unmatched_tasks(unmatched_tasks)
            sys.exit(1)

    def _match_task(self, task: BaseTask) -> Optional[AdhesiveBaseTask]:
        for step in self.process.steps:
            if step.matches(task, task.name) is not None:
                return step

        return None

    def clone_event(self,
                    old_event: ActiveEvent,
                    task: BaseTask,
                    parent_id: Optional[str] = None) -> ActiveEvent:

        if parent_id is None:
            parent_id = old_event.parent_id

        event = old_event.clone(task, parent_id)
        self.register_event(event)

        if parent_id is None:
            workflow = self.process.workflow
        else:
            workflow = self.events[parent_id].task

        def process_event(_event) -> ActiveEventState:
            # if there is no processing needed, we skip to routing
            if isinstance(event.task, StartEvent) or isinstance(event.task, EndEvent):
                return event.state.route(event.context)

            if isinstance(event.task, NonWaitingGateway):
                return event.state.route(event.context)

            # if we need to wait, we wait.
            if isinstance(event.task, WaitingGateway):
                return event.state.wait_check()

            # normally we shouldn't wait for tasks, since it's counter BPMN, so
            # we allow configuring waiting for it.
            if self.config.wait_tasks and (
                    isinstance(event.task, ScriptTask) or
                    isinstance(event.task, UserTask) or
                    isinstance(event.task, Task) or
                    isinstance(event.task, Workflow)
            ):
                return event.state.wait_check()

            return event.state.run()

        def get_other_task_waiting(source: ActiveEvent) -> \
                Tuple[Optional[ActiveEvent], int]:
            """
            Get any other task that is waiting to be executed on the
            same task that's waiting.
            :param source:
            :return:
            """
            result = None
            count = 0

            for ev in self.events.values():
                if ev == source:
                    continue

                if ev.task == source.task and \
                        ev.state.state in PRE_RUN_STATES and \
                        loop_id(ev) == loop_id(source):
                    if not result:
                        result = ev

                    count += 1

            return result, count

        def wait_task(_event) -> Optional[ActiveEventState]:
            # is another waiting task already present?
            other_waiting, tasks_waiting_count = get_other_task_waiting(event)

            potential_predecessors = list(map(
                lambda e: e.task,
                filter(lambda e: is_predecessor(event, e), self.events.values())))

            if other_waiting:
                new_data = WorkflowData.merge(other_waiting.context.data, event.context.data)
                other_waiting.context.data = new_data

                event.state.done()

            # if we have predecessors, we stay in waiting
            if workflow.are_predecessors(event.task, potential_predecessors):
                return None

            if not other_waiting:
                return event.state.run()

            if other_waiting.state.state == ActiveEventState.WAITING and \
                    tasks_waiting_count == 1:
                other_waiting.state.run()

            return None

        def run_task(_event) -> None:
            # If the event is not yet started as a loop, we need to do that. We might have a wrong
            # loop context, if we're running in a nested loop.
            if event.task.loop and (not event.context.loop or event.context.loop.task != event.task):
                # we start a loop by firing the loop events, and consume this event.
                WorkflowLoop.create_loop(event, self.clone_event)
                event.state.done_check(None)

                return

            if isinstance(event.task, Workflow):
                for start_task in event.task.start_tasks.values():
                    # this automatically registers our events for execution
                    self.clone_event(event, start_task, parent_id=event.id)
                return

            if isinstance(event.task, Task):
                future = WorkflowExecutor.pool.submit(
                    self.tasks_impl[event.task.id].invoke,
                    event.context)
                self.futures[future] = event.id
                event.future = future
                return

            if isinstance(event.task, ScriptTask):
                future = WorkflowExecutor.pool.submit(
                    call_script_task,
                    event)
                self.futures[future] = event.id
                event.future = future
                return

            if isinstance(event.task, UserTask):
                future = Future()
                self.futures[future] = event.id
                event.future = future

                self.ut_provider.register_event(self, event)

                return

            return event.state.route(event.context)

        def error_task(_event) -> None:
            # if we have a boundary error task, we use that one for processing.
            if event.task.error_task:
                self.clone_event(event, event.task.error_task)

                if event.task.error_task.cancel_activity:
                    event.state.done_check(None)
                else:
                    event.state.route()

                return

            if event.parent_id in self.events:
                parent_event = self.get_parent(event.id)

                # we move the parent into error
                parent_event.state.error(_event.data)

                # FIXME: not sure if this is enough.
                for potential_child in list(self.events.values()):
                    if potential_child.parent_id != event.parent_id:
                        continue

                    # potential_child.state.error(_event.data)
                    potential_child.state.done()

            # event.state.done_check(None)  # we kill the current event

        def route_task(_event) -> None:
            event.context = _event.data
            outgoing_edges = GatewayController.compute_outgoing_edges(workflow, event)

            for outgoing_edge in outgoing_edges:
                target_task = workflow.tasks[outgoing_edge.target_id]
                self.clone_event(event, target_task)

            event.state.done_check(outgoing_edges)

        def done_check(_event) -> None:
            outgoing_edges = _event.data

            if not outgoing_edges:
                return event.state.done_end_task()

            return event.state.done()

        def done_end_task(_event) -> None:
            # we should check all the WAITING processes if they finished.
            event_count: Dict[BaseTask, int] = dict()
            waiting_events: List[ActiveEvent] = list()

            for id, self_event in self.events.items():
                if self_event.state.state in DONE_STATES:
                    continue

                event_count[self_event.task] = event_count.get(self_event.task, 0) + 1

                if self_event.state.state == ActiveEventState.WAITING:
                    waiting_events.append(self_event)

            for waiting_event in waiting_events:
                if event_count[waiting_event.task] > 1:
                    continue

                potential_predecessors = list(map(
                    lambda e: e.task,
                    filter(lambda e: is_predecessor(waiting_event, e), self.events.values())))

                if not workflow.are_predecessors(waiting_event.task, potential_predecessors):
                    waiting_event.state.run()

            # check sub-process termination
            found = False
            for ev in self.events.values():
                if ev.parent_id == event.parent_id and ev != event:
                    found = True
                    break

            # we merge into the parent event if it's an end state.
            if parent_id is not None:
                self.events[parent_id].context.data = WorkflowData.merge(
                    self.events[parent_id].context.data,
                    event.context.data
                )

                if not found:
                    parent_event = self.events[parent_id]
                    parent_event.state.route(parent_event.context)

            event.state.done()

        def done_task(_event) -> None:
            # if the task is done, but the future is still registered, we need to
            # remove the future.
            if event.future in self.futures:
                del self.futures[event.future]

            self.unregister_event(event)

        event.state.after_enter(ActiveEventState.PROCESSING, process_event)
        event.state.after_enter(ActiveEventState.WAITING, wait_task)
        event.state.after_enter(ActiveEventState.RUNNING, run_task)
        event.state.after_enter(ActiveEventState.ERROR, error_task)
        event.state.after_enter(ActiveEventState.ROUTING, route_task)
        event.state.after_enter(ActiveEventState.DONE_CHECK, done_check)
        event.state.after_enter(ActiveEventState.DONE_END_TASK, done_end_task)
        event.state.after_enter(ActiveEventState.DONE, done_task)

        return event



import os
import uuid
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Optional


class Workspace(ABC):
    """
    A workspace is a place where work can be done. That means a writable
    folder is being allocated, that will be cleaned up at the end of the
    execution.
    """
    def __init__(self,
                 pwd: str) -> None:
        self.pwd = pwd

    @abstractmethod
    def write_file(
            self,
            file_name: str,
            content: str) -> None:
        pass

    @abstractmethod
    def run(self, command: str) -> None:
        """
        Run a new command in the current workspace.

        :param command:
        :return:
        """
        pass

    @abstractmethod
    def rm(self, path: Optional[str]=None) -> None:
        """
        Recursively remove the file or folder given as path. If no path is sent,
        the whole workspace will be cleared.

        :param path:
        :return:
        """
        pass

    @abstractmethod
    def mkdir(self, path: str=None) -> None:
        """
        Create a folder, including all its needed parents.

        :param path:
        :return:
        """
        pass

    @abstractmethod
    def copy_to_agent(self,
                      from_path: str,
                      to_path: str) -> None:
        """
        Copy the files to the agent from the current disk.
        :param from_path:
        :param to_path:
        :return:
        """
        pass

    @abstractmethod
    def copy_from_agent(self,
                        from_path: str,
                        to_path: str) -> None:
        """
        Copy the files from the agent to the current disk.
        :param from_path:
        :param to_path:
        :return:
        """
        pass

    @contextmanager
    def temp_folder(self):
        """
        Create a temporary folder in the current `pwd` that will be deleted
        when the `with` block ends.

        :return:
        """
        current_folder = self.pwd
        folder = os.path.join(self.pwd, str(uuid.uuid4()))

        self.mkdir(folder)
        self.pwd = folder

        try:
            yield folder
        finally:
            self.rm(folder)
            self.pwd = current_folder

    @contextmanager
    def chdir(self, target_folder: str):
        current_folder = self.pwd
        folder = os.path.join(self.pwd, target_folder)

        self.pwd = folder

        try:
            yield folder
        finally:
            self.pwd = current_folder



from typing import Callable, List, Optional

from adhesive.graph.BaseTask import BaseTask
from adhesive.graph.Task import Task
from adhesive.steps.AdhesiveBaseTask import AdhesiveBaseTask
from .WorkflowContext import WorkflowContext


class AdhesiveTask(AdhesiveBaseTask):
    """
    A task implementation.
    """
    def __init__(self,
                 code: Callable,
                 *expressions: str) -> None:
        super(AdhesiveTask, self).__init__(code, *expressions)

    def matches(self,
                task: BaseTask,
                resolved_name: str) -> Optional[List[str]]:
        if not isinstance(task, Task):
            return None

        return super(AdhesiveTask, self).matches(task, resolved_name)

    def invoke(
            self,
            context: WorkflowContext) -> WorkflowContext:
        params = self.matches(context.task, context.task_name)

        self.code(context, *params)  # type: ignore

        return context



from adhesive.graph.BaseTask import BaseTask


class ScriptTask(BaseTask):
    def __init__(self,
                 _id: str,
                 name: str,
                 language: str,
                 script: str) -> None:
        super(ScriptTask, self).__init__(_id, name)

        self.language = language
        self.script = script



from typing import Iterable

import re

from adhesive.graph.BaseTask import BaseTask
from adhesive.graph.UserTask import UserTask


def display_unmatched_tasks(unmatched_tasks: Iterable[BaseTask]) -> None:
    print("Missing tasks implementations. Generate with:\n")

    for unmatched_task in unmatched_tasks:
        if isinstance(unmatched_task, UserTask):
            print(f"@adhesive.usertask('{re.escape(unmatched_task.name)}')")
            print("def task_impl(context, ui):")
            print("    pass\n\n")
            continue

        print(f"@adhesive.task('{re.escape(unmatched_task.name)}')")
        print("def task_impl(context):")
        print("    pass\n\n")



import distutils
import os
import shutil
import subprocess
import uuid
from typing import Optional
from distutils.dir_util import copy_tree

from adhesive.workspace.Workspace import Workspace
import logging

LOG = logging.getLogger(__name__)


class LocalLinuxWorkspace(Workspace):
    """
    A workspace is a place where work can be done. That means a writable
    folder is being allocated, that will be cleaned up at the end of the
    execution.
    """
    def __init__(self,
                 pwd: Optional[str]=None) -> None:
        if not pwd:
            # FIXME: detect the temp folder
            pwd = os.path.join("/tmp/", str(uuid.uuid4()))
            os.mkdir(pwd)

        super(LocalLinuxWorkspace, self).__init__(pwd=pwd)

    def run(self, command: str) -> None:
        subprocess.check_call([
            "/bin/sh", "-c", command
        ], cwd=self.pwd)

    def write_file(
            self,
            file_name: str,
            content: str) -> None:

        full_path = os.path.join(self.pwd, file_name)

        with open(full_path, "wt") as f:
            f.write(content)

    def rm(self, path: Optional[str]=None) -> None:
        if path is None:
            LOG.debug("rmtree {}", self.pwd)
            shutil.rmtree(self.pwd)
            return

        if not path:
            raise Exception("You need to pass a subpath to delete")

        remove_path = os.path.join(self.pwd, path)

        LOG.debug("rmtree {}", remove_path)

        if os.path.isfile(remove_path):
            os.remove(remove_path)
        else:
            shutil.rmtree(remove_path)

    def mkdir(self, path: str=None) -> None:
        LOG.debug("mkdir {}", path)
        os.mkdir(os.path.join(self.pwd, path))

    def copy_to_agent(self,
                      from_path: str,
                      to_path: str):
        LOG.debug("copy {} to {}", from_path, to_path)
        copy_tree(from_path, to_path)

    def copy_from_agent(self,
                        from_path: str,
                        to_path: str):
        LOG.debug("copy {} to {}", from_path, to_path)
        shutil.copytree(from_path, to_path)



import unittest

from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.xml.bpmn import read_bpmn_file

from test.adhesive.steps.check_equals import assert_equal_steps
from test.adhesive.steps.test_tasks import adhesive, _async


class TestLoopExecution(unittest.TestCase):

    def test_loop_execution(self):
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/loop.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process)
        data = _async(workflow_executor.execute())

        assert_equal_steps({
            'Build Germanium Image on mac': 1,
            'Build Germanium Image on windows': 1,
            'Build Germanium Image on linux': 1,
            'Test Browser ie on linux': 1,
            'Cleanup Platform linux': 1,
            'Cleanup Platform windows': 1,
            'Cleanup Platform mac': 1,
            'Test Browser chrome on linux': 1,
            'Test Browser edge on linux': 1,
            'Test Browser edge on windows': 1,
            'Test Browser chrome on windows': 1,
            'Test Browser ie on windows': 1,
            'Test Browser chrome on mac': 1,
            'Test Browser edge on mac': 1,
            'Test Browser ie on mac': 1,
        }, data.steps)
        self.assertFalse(workflow_executor.events)

    def test_loop_execution_no_wait(self):
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/loop.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process, wait_tasks=False)
        data = _async(workflow_executor.execute())

        assert_equal_steps({
            'Build Germanium Image on mac': 1,
            'Build Germanium Image on windows': 1,
            'Build Germanium Image on linux': 1,
            'Test Browser ie on linux': 1,
            'Cleanup Platform linux': 3,
            'Cleanup Platform windows': 3,
            'Cleanup Platform mac': 3,
            'Test Browser chrome on linux': 1,
            'Test Browser edge on linux': 1,
            'Test Browser edge on windows': 1,
            'Test Browser chrome on windows': 1,
            'Test Browser ie on windows': 1,
            'Test Browser chrome on mac': 1,
            'Test Browser edge on mac': 1,
            'Test Browser ie on mac': 1,
        }, data.steps)
        self.assertFalse(workflow_executor.events)


if __name__ == '__main__':
    unittest.main()



import subprocess
from contextlib import contextmanager
from typing import Optional, Union, Iterable

from .Workspace import Workspace


class DockerWorkspace(Workspace):
    def __init__(self,
                 image_name: str,
                 pwd: str) -> None:
        super(DockerWorkspace, self).__init__(pwd)
        self.container_id = subprocess.check_output([
            "docker", "run", "-t", "-v", f"{pwd}:{pwd}", "-d", "--entrypoint", "cat", image_name
        ]).decode('utf-8').strip()

    def run(self, command: str) -> None:
        subprocess.check_call([
            "docker", "exec", "-w", self.pwd, self.container_id, "/bin/sh", "-c", command
        ], cwd=self.pwd)

    def write_file(
            self,
            file_name: str,
            content: str) -> None:
        raise Exception("not implemented")

    def rm(self, path: Optional[str]=None) -> None:
        raise Exception("not implemented")

    def mkdir(self, path: str=None) -> None:
        raise Exception("not implemented")

    def copy_to_agent(self,
                      from_path: str,
                      to_path: str):
        subprocess.check_call([
            "docker", "cp", from_path, f"{self.container_id}:{to_path}"
        ])

    def copy_from_agent(self,
                        from_path: str,
                        to_path: str):
        subprocess.check_call([
            "docker", "cp", f"{self.container_id}:{from_path}", to_path
        ])

    def _destroy(self):
        subprocess.check_call([
            "docker", "rm", "-f", self.container_id
        ])


@contextmanager
def inside(workspace: Workspace,
           image_name: str):
    w = None

    try:
        w = DockerWorkspace(image_name, pwd=workspace.pwd)
        yield w
    finally:
        if w is not None:
            w._destroy()


@contextmanager
def build(workspace: Workspace,
          tags: Union[str, Iterable[str]]):

    # we always consider a list of tags
    if isinstance(tags, str):
        tags = [tags]

    command = "docker build "

    for tag in tags:
        command += f"-t {tag} "

    command += "."

    workspace.run(command)



import unittest

from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.xml.bpmn import read_bpmn_file

from .test_tasks import adhesive, _async
from .check_equals import assert_equal_steps


class TestScriptTask(unittest.TestCase):
    """
    Test if the workflow executor can process script tasks.
    """
    def test_script_task_execution(self):
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/script.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process)
        data = _async(workflow_executor.execute())

        assert_equal_steps({
            'Script Task': 1,
            'Build Germanium Image': 1,
        }, data.steps)

        self.assertFalse(workflow_executor.events,
                         "Some events were not unregistered and/or executed.")



from typing import Optional, List, Dict, Any, cast, Tuple, Union, Iterable

import npyscreen as npyscreen

from adhesive.model.UiBuilderApi import UiBuilderApi
from adhesive.steps.AdhesiveUserTask import AdhesiveUserTask
from adhesive.model.ActiveEvent import ActiveEvent
from adhesive.model.UserTaskProvider import UserTaskProvider
from adhesive.model.WorkflowExecutor import WorkflowExecutor

from adhesive.steps.WorkflowData import WorkflowData


class UIBuilder(UiBuilderApi):
    def __init__(self,
                 event: ActiveEvent):
        self.context = event.context
        self.ui_controls: Dict[str, Any] = dict()

        self.labels: Dict[str, List[str]] = dict()
        self.values: Dict[str, List[str]] = dict()

        self.form = npyscreen.Form(
            name=event.task.name)

    @property
    def data(self) -> WorkflowData:
        result_dict = dict()

        for name, ui_control in self.ui_controls.items():
            if isinstance(self.ui_controls[name], npyscreen.TitleMultiSelect):
                result = {self.values[name][it] for it in ui_control.value}
                result_dict[name] = result
                continue

            if isinstance(self.ui_controls[name], npyscreen.TitleCombo):
                result = self.values[name][ui_control.value] if ui_control.value >= 0 else None
                result_dict[name] = result
                continue

            if isinstance(self.ui_controls[name], npyscreen.TitleSelectOne):
                result = self.values[name][ui_control.value[0]] if len(ui_control.value) > 0 else None
                result_dict[name] = result
                continue

            result_dict[name] = ui_control.value

        return WorkflowData(result_dict)

    def add_input_text(self,
                       name: str,
                       title: Optional[str] = None,
                       value: str = '') -> None:
        if not name:
            raise Exception("You need to pass a name for the input.")

        if not title:
            title = name

        self.ui_controls[name] = self.form.add_widget(
            npyscreen.TitleText,
            name=title,
            value=value)

    def add_input_password(self,
                       name: str,
                       title: Optional[str] = None,
                       value: str = '') -> None:
        if not name:
            raise Exception("You need to pass a name for the input.")

        if not title:
            title = name

        self.ui_controls[name] = self.form.add_widget(
            npyscreen.TitlePassword,
            name=title,
            value=value)

    def add_combobox(
            self,
            name: str,
            title: Optional[str] = None,
            value: Optional[str]=None,
            values: Optional[Iterable[Union[Tuple[str, str], str]]]=None) -> None:
        if not name:
            raise Exception("You need to pass a name for the input.")

        if not title:
            title = name

        self.values[name] = UIBuilder._get_values(values)
        self.labels[name] = UIBuilder._get_labels(values)

        _value = self.values[name].index(UIBuilder._get_value(value)) if value else -1

        self.ui_controls[name] = self.form.add_widget(
            npyscreen.TitleCombo,
            name=title,
            value=_value,
            values=self.labels[name])

    def add_checkbox_group(
            self,
            name: str,
            title: Optional[str]=None,
            value: Optional[Iterable[str]]=None,
            values: Optional[Iterable[Union[Tuple[str, str], str]]]=None) -> None:
        if not name:
            raise Exception("You need to pass a name for the input.")

        if not values:
            raise Exception("You need to pass in some values to display in the group.")

        if not title:
            title = name

        self.values[name] = UIBuilder._get_values(values)
        self.labels[name] = UIBuilder._get_labels(values)

        self.ui_controls[name] = self.form.add_widget(
            npyscreen.TitleMultiSelect,
            name=title,
            value=[self.values[name].index(UIBuilder._get_value(v)) for v in value],
            max_height=max(len(values) + 1, 2),
            scroll_exit=True,
            values=self.labels[name])

    def add_radio_group(self,
                        name: str,
                        title: Optional[str]=None,
                        value: Optional[str]=None,
                        values: Optional[Iterable[Union[Tuple[str, str], str]]]=None) -> None:
        if not name:
            raise Exception("You need to pass a name for the input.")

        if not title:
            title = name

        self.values[name] = UIBuilder._get_values(values)
        self.labels[name] = UIBuilder._get_labels(values)

        _value = self.values[name].index(UIBuilder._get_value(value)) if value else -1

        self.ui_controls[name] = self.form.add_widget(
            npyscreen.TitleSelectOne,
            name=title,
            max_height=max(len(values), 2),
            scroll_exit=True,
            value=_value,
            values=self.labels[name])

    def add_default_button(self,
                           name: str,
                           title: Optional[str]=None) -> None:
        self.context.data[name] = name

    @staticmethod
    def _get_values(values) -> List[str]:
        result = []

        for value in values:
            if isinstance(value, str):
                result.append(value)
                continue

            result.append(value[0])

        return result

    @staticmethod
    def _get_value(value) -> str:
        if isinstance(value, str):
            return value

        return value[0]

    @staticmethod
    def _get_labels(values) -> List[str]:
        result = []

        for value in values:
            if isinstance(value, str):
                result.append(value)
                continue

            result.append(value[1])

        return result


class ConsoleUserTaskProvider(UserTaskProvider):
    def __init__(self):
        super(ConsoleUserTaskProvider, self).__init__()

    def register_event(self,
                       executor: WorkflowExecutor,
                       event: ActiveEvent) -> None:
        def run_on_curses(x):
            try:
                ui = UIBuilder(event)

                adhesive_task = cast(AdhesiveUserTask, executor.tasks_impl[event.task.id])
                context = adhesive_task.invoke_user_task(event.context, ui)

                # call the actual UI
                ui.form.edit()

                # put the values in the data.
                context.data = WorkflowData.merge(context.data, ui.data)

                event.future.set_result(context)
            except Exception as e:
                event.future.set_exception(e)

        npyscreen.wrapper_basic(run_on_curses)



import adhesive
import time
import asyncio
import uuid


def _async(fn):
    data = asyncio.get_event_loop().run_until_complete(fn)
    return data


@adhesive.task(
    'Ensure Docker Tooling',
    'Test Chrome',
    'Test Firefox',
    'Build Germanium Image',
    'Prepare Firefox',
    # exclusive gateway
    'Exclusive\ Task\ Branch',
    'Populate\ task\ data',
    'Exclusive\ default\ branch',
    'Cleanup Broken Tasks',
    'Error Was Caught',
    'Error Was Not Caught',
    '^Cleanup Platform .*?$',
    '^Test Browser .*? on .*?$'
)
def basic_task(context) -> None:
    add_current_task(context)


@adhesive.task(r'^Parallel \d+$')
def parallel_task(context) -> None:
    time.sleep(1)
    if not context.data.steps:
        context.data.steps = set()

    context.data.steps.add(context.task_name)


@adhesive.task(
    r'^Throw Some Exception$',
    'Throw Some Error',
)
def throw_some_exception(context) -> None:
    add_current_task(context)

    raise Exception("broken")


@adhesive.task('Increment\ X\ by\ 1')
def increment_x_by_1(context):
    add_current_task(context)

    if not context.data.x:
        context.data.x = 1
        return

    context.data.x += 1


@adhesive.usertask('Read Data From User')
def read_data_from_user(context, ui) -> None:
    ui.add_input_text("branch", title="Branch")
    ui.add_input_password("password", title="Password")
    ui.add_combobox("version", title="Version", values=["12.0", "12.1", "12.2", "12.3"])
    ui.add_checkbox_group(
        "run_tests",
        title="Tests",
        value=("integration",),
        values=("integration", "Integration Tests"))
    ui.add_radio_group(
        "depman",
        title="Depman"
    )

    ui.add_default_button("OK")
    ui.add_default_button("Cancel")


def add_current_task(context):
    if not context.data.steps:
        context.data.steps = dict()

    if context.task_name not in context.data.steps:
        context.data.steps[context.task_name] = set()

    context.data.steps[context.task_name].add(str(uuid.uuid4()))



from typing import Optional, Dict, Any
from adhesive.steps.WorkflowData import WorkflowData
from adhesive.graph.BaseTask import BaseTask
from adhesive.workspace.Workspace import Workspace
from adhesive.workspace.local.LocalLinuxWorkspace import LocalLinuxWorkspace


class WorkflowContext:
    """
    A context passed to an execution of a task. It holds the information
    about:
    - data that's being populated across tasks,
    - workspace where files can be created.
    """
    def __init__(self,
                 task: 'BaseTask',
                 data: Optional[Dict]=None,
                 workspace: Optional[Workspace]=None) -> None:
        self.task = task
        self.data = WorkflowData(data)
        self.workspace: Workspace = LocalLinuxWorkspace() if not workspace else workspace
        self.loop: Optional[WorkflowLoop] = None

        self.update_title()

    def update_title(self) -> None:
        # FIXME: this breaks the encapsulation of the data
        try:
            self.task_name = self.task.name.format(**{
                "context": self,
                "data": self.data,
                "loop": self.loop
            })
        except Exception as e:
            self.task_name = self.task.name

    def clone(self, task: 'BaseTask') -> 'WorkflowContext':
        result = WorkflowContext(
            task,
            self.data.as_dict(),
            self.workspace,  # FIXME: this should be a clone
        )

        return result

    def as_mapping(self) -> Dict[str, Any]:
        return {
            "task": self.task,
            "data": self.data,
            "loop": self.loop,
            "task_name": self.task_name,
        }


from adhesive.steps.WorkflowLoop import WorkflowLoop
